/* 
 * btmvbe.c --
 *
 *	Boundary Tracing Generation Method, traces the outline of areas
 *      of a single color and fills them in.
 *
 * Copyright (c) 1994-1997 Michael R. Ganss. All Rights Reserved.
 *
 * This file is distributed under the same conditions as
 * AlmondBread. For further information see
 * <URL:http://www.cs.tu-berlin.de/~rms/AlmondBread>. 
 *
 */
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#undef WORDS_BIGENDIAN
#include "ieee754.h"
#include "vbe.h"
#include "default.h"

#define TRUE 1
#define FALSE 0
#define EVERY 15
#define BASIN_COLOR 0
#define INTEL

static double stepx, stepy;
static double minr=-2.5, maxr=1.5, mini=-1.5, maxi=1.5;
static double basx, basy;
static float equal;
static int anz=152;
static char basin = FALSE;
static int maxx, maxy, midy, miny;
static int y;
static int verbose=FALSE;
extern int optind,opterr;
extern char *optarg;

static unsigned char adjust[256]={0,1,2,2,3,3,3,3,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8};
				  
int iteration(double re, double im)
{
  register int iter,k,n;
  register double ren,imn,cr,ci,sre,sim;
#ifdef INTEL
  register float mag;
  register unsigned long bail=0x41800000,magi; /* bail=16.0 */
#if 0 /* see below for explanation */
  register unsigned long eq=*(unsigned long *)&equal;
#endif
#else
  register double mag;
#endif
  union ieee754_double d;
  int exponent;
  
  if(basin)
    {
      cr=sre=re; ci=sim=im;
      ren=re*re;
      imn=im*im;
      iter=anz>>3;
      k=n=8;
      
      do
	{
	  im=im*re;
	  re=ren-imn;
	  im+=im;
	  re+=cr;
	  im+=ci;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;
	  
#if 0 /* this doesn't work properly, because integer compares are only
	 32 bits */
	  mag=fabs(sre-re);
	  magi=*(unsigned long *)&mag;
	  if(magi<eq)
	    {
	      mag=fabs(sim-im);
	      magi=*(unsigned long *)&mag;
	      if(magi<eq)
		return BASIN_COLOR;
	    }
#else
	  if(fabs(sre-re)<equal&&fabs(sim-im)<equal)
	    return BASIN_COLOR;
#endif

	  k-=8;
	  if(k<=0)
	    {
	      n<<=1;
	      sre=re;
	      sim=im;
	      k=n;
	    }

	  imn=im*im;
	  ren=re*re;
	  mag=ren+imn;
#ifdef INTEL
	  magi=*(unsigned long *)&mag;
#endif
	}
#ifdef INTEL
      while (magi<bail&&--iter!=0);
#else
      while (mag<16.0&&--iter!=0);
#endif
    }
  else
    {
      cr=re; ci=im;
      ren=re*re;
      imn=im*im;
      iter=anz>>3;

      do
	{      
	  im=im*re;
	  re=ren-imn;
	  im+=im;
	  re+=cr;
	  im+=ci;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*im;
	  ren=re*re;
	  mag=ren+imn;
#ifdef INTEL
	  magi=*(unsigned long *)&mag;
#endif
	}
#ifdef INTEL
      while (magi<bail&&--iter!=0);
#else
      while (mag<16.0&&--iter!=0);
#endif
    }
  
  if(iter==0)
    {
      basin=TRUE;
      return BASIN_COLOR;
    }
  else
    {
      basin=FALSE;
#ifdef INTEL
      d.d=ren+imn;
#else
      d.d=mag;
#endif
      exponent=(d.ieee.exponent-_IEEE754_DOUBLE_BIAS)>>3;

      return (anz-(((iter-1)<<3)+adjust[exponent]));
    }
}

#define RIGHT 0
#define LEFT 1
#define UP 2
#define DOWN 3

static int **screen;
/* Boundary Tracing Method
   
   We start from pixel (x,y) representing (re,im) going right and
   keeping the "wall" to our left until we get back to the starting
   point. For each pixel, the next pixel visited will be computed
   thus:
   
   a) check pixel to left (looking in the current direction, i.e. if
   the current direction is right, the pixel to the left is the one
   above it). If it's part of the wall (i.e. offscreen or a different
   color than the one we're tracking), go to b). If it's not, it is
   the next pixel.
   b) do same for pixel straight ahead. Go there if it's not part of
   the wall.
   c) ditto for pixel to the right. If this pixel is also part of the
   wall, go back one pixel.
   
   When we're back at the starting point we trace the boundary once
   more. Now, whenever we move up, we plot pixels to the right of the
   current one until we hit the wall.

   This method was inspired by a post on sci.fractals by Maarten
   Egmond (cp@stack.urc.tue.nl).
   */
void trace(int x, int y, double re, double im)
{
  int sx,sy,c,w,dir=RIGHT,fill=0,x1,color;
  int *s;
  
  c=iteration(re,im); /* the color we're tracing */
  color=c&255;

  sx=x; sy=y;

  do
    {
      if(!fill&&screen[y][x]<0) /* plot current pixel */
	{
	  screen[y][x]=c;
	  putpixel(x,y,color);
	}
      
      switch(dir)
	{
	case RIGHT:
	  /* check left */
	  if(y-1>=miny)
	    {
	      if((w=screen[y-1][x])<0)
		{
		  w=iteration(re,im-stepy);
		  screen[y-1][x]=w;
		  putpixel(x,y-1,w&255);
		}
	      if(w==c) /* go left */
		{
		  dir=UP;
		  im-=stepy;
		  y--;
		  break;
		}
	    }
	  /* check ahead */
	  if(x+1<maxx)
	    {
	      if((w=screen[y][x+1])<0)
		{
		  w=iteration(re+stepx,im);
		  screen[y][x+1]=w;
		  putpixel(x+1,y,w&255);
		}
	      if(w==c) /* go ahead */
		{
		  /* dir is still RIGHT */
		  re+=stepx;
		  x++;
		  break;
		}
	    }
	  /* check right */
	  if(y+1<maxy)
	    {
	      if((w=screen[y+1][x])<0)
		{
		  w=iteration(re,im+stepy);
		  screen[y+1][x]=w;
		  putpixel(x,y+1,w&255);
		}
	      if(w==c) /* go right */
		{
		  dir=DOWN;
		  im+=stepy;
		  y++;
		  break;
		}
	    }
	  /* go back */
	  if(x==sx&&y==sy) /* if direction is RIGHT and we're at the */
			   /* starting point, going back means return */
	    return;
	  dir=LEFT;
	  re-=stepx;
	  x--;
	  break;
	  
	case LEFT:
	  /* check left */
	  if(y+1<maxy)
	    {
	      if((w=screen[y+1][x])<0)
		{
		  w=iteration(re,im+stepy);
		  screen[y+1][x]=w;
		  putpixel(x,y+1,w&255);
		}
	      if(w==c) /* go left */
		{
		  dir=DOWN;
		  im+=stepy;
		  y++;
		  break;
		}
	    }
	  /* check ahead */
	  if(x-1>=0)
	    {
	      if((w=screen[y][x-1])<0)
		{
		  w=iteration(re-stepx,im);
		  screen[y][x-1]=w;
		  putpixel(x-1,y,w&255);
		}
	      if(w==c) /* go ahead */
		{
		  /* dir is still LEFT */
		  re-=stepx;
		  x--;
		  break;
		}
	    }
	  /* check right */
	  if(y-1>=miny)
	    {
	      if((w=screen[y-1][x])<0)
		{
		  w=iteration(re,im-stepy);
		  screen[y-1][x]=w;
		  putpixel(x,y-1,w&255);
		}
	      if(w==c) /* go right */
		{
		  dir=UP;
		  im-=stepy;
		  y--;
		  break;
		}
	    }
	  /* go back */
	  dir=RIGHT;
	  re+=stepx;
	  x++;
	  break;
	  
	case UP:
	  if(fill) /* fill current line by going left until we hit a */
		   /* pixel with a different color than c */
	    {
	      s=&screen[y][maxx];
	      x1=-maxx+x+1;
	      for(; x1!=0; x1++)
		{
		  if(s[x1]<0)
		    s[x1]=c;
		  else
		    if(s[x1]!=c)
		      break;
		}
	      x1+=maxx;
	      if(x1-2>x)
		puthline(x+1,y,x1-1,color);
	    }
	  /* check left */
	  if(x-1>=0)
	    {
	      if((w=screen[y][x-1])<0)
		{
		  w=iteration(re-stepx,im);
		  screen[y][x-1]=w;
		  putpixel(x-1,y,w&255);
		}
	      if(w==c) /* go left */
		{
		  dir=LEFT;
		  re-=stepx;
		  x--;
		  break;
		}
	    }
	  /* check ahead */
	  if(y-1>=miny)
	    {
	      if((w=screen[y-1][x])<0)
		{
		  w=iteration(re,im-stepy);
		  screen[y-1][x]=w;
		  putpixel(x,y-1,w&255);
		}
	      if(w==c) /* go ahead */
		{
		  /* dir is still UP */
		  im-=stepy;
		  y--;
		  break;
		}
	    }
	  /* check right */
	  if(x+1<maxx)
	    {
	      if((w=screen[y][x+1])<0)
		{
		  w=iteration(re+stepx,im);
		  screen[y][x+1]=w;
		  putpixel(x+1,y,w&255);
		}
	      if(w==c) /* go right */
		{
		  dir=RIGHT;
		  re+=stepx;
		  x++;
		  break;
		}
	    }
	  /* go back */
	  dir=DOWN;
	  im+=stepy;
	  y++;
	  break;
	  
	case DOWN:
	  /* check left */
	  if(x+1<maxx)
	    {
	      if((w=screen[y][x+1])<0)
		{
		  w=iteration(re+stepx,im);
		  screen[y][x+1]=w;
		  putpixel(x+1,y,w&255);
		}
	      if(w==c) /* go left */
		{
		  dir=RIGHT;
		  re+=stepx;
		  x++;
		  break;
		}
	    }
	  /* check ahead */
	  if(y+1<maxy)
	    {
	      if((w=screen[y+1][x])<0)
		{
		  w=iteration(re,im+stepy);
		  screen[y+1][x]=w;
		  putpixel(x,y+1,w&255);
		}
	      if(w==c) /* go ahead */
		{
		  /* dir is still DOWN */
		  im+=stepy;
		  y++;
		  break;
		}
	    }
	  /* check right */
	  if(x-1>=0)
	    {
	      if((w=screen[y][x-1])<0)
		{
		  w=iteration(re-stepx,im);
		  screen[y][x-1]=w;
		  putpixel(x-1,y,w&255);
		}
	      if(w==c) /* go right */
		{
		  dir=LEFT;
		  re-=stepx;
		  x--;
		  break;
		}
	    }
	  /* go back */
	  dir=UP;
	  im-=stepy;
	  y--;
	  break;
	}

      if(x==sx&&y==sy) /* we're back where we started */
	{
	  if(fill) /* we're done */
	    return;
	  fill=1;  /* trace again, filling encircled area */
	  dir=RIGHT;
	}
    }
  while(1);
}

void LoadColorMap(char *filename)
{
  FILE *f;
  int n=256,index,r,g,b;
  char c;

  if((f=fopen(filename,"r"))==NULL)
    {
      perror(filename);
      close_graphics();
      exit(1);
    }
  for(index=0; index<n; index++)
    {
      if(fscanf(f,"%d",&r)==EOF)
	{
	  if(ferror(f))
	    perror(filename);
	  else fprintf(stderr,"%s: Premature EOF.\n",filename);
	  close_graphics();
	  exit(1);
	}
      if(fscanf(f,"%d",&g)==EOF)
	{
	  if(ferror(f))
	    perror(filename);
	  else fprintf(stderr,"%s: Premature EOF.\n",filename);
	  close_graphics();
	  exit(1);
	}
      if(fscanf(f,"%d",&b)==EOF)
	{
	  if(ferror(f))
	    perror(filename);
	  else fprintf(stderr,"%s: Premature EOF.\n",filename);
	  close_graphics();
	  exit(1);
	}
      while(((c=fgetc(f))!='\n')&&(c!=EOF));
      set_palette(index,(int)((double)r/4+0.5),
		  (int)((double)g/4+0.5),(int)((double)b/4+0.5));
    }
  fclose(f);
}

void usage(char **argv)
{
  int i,n;
  
  fprintf(stderr,"Usage: %s %n[-v]\n",argv[0],&n);
  for(i=0; i<n; i++)
    fprintf(stderr," ");
  fprintf(stderr,"[-i max # of iterations]\n");
  for(i=0; i<n; i++)
    fprintf(stderr," ");
  fprintf(stderr,"[-c colormap]\n");
  for(i=0; i<n; i++)
    fprintf(stderr," ");
  fprintf(stderr,"[-m graphics mode 1|2|3]\n");
  for(i=0; i<n; i++)
    fprintf(stderr," ");
  fprintf(stderr,"-- [minr] [maxr] [mini] [maxi]\n");
  
  exit(1);
}

void Zoom(void)
{
  int x,y,button;
  double cr,ci;

  vbe_cursor=0;
  
  while(1)
    {
      mouse_position(&x,&y,&button);
      display_cursor(x,y);
      if(button&1)
	{
	  cr=minr+x*stepx;
	  ci=mini+y*stepy;
	  minr=cr-(vbe_width/4)*stepx;
	  maxr=cr+(vbe_width/4)*stepx;
	  mini=ci-(vbe_height/4)*stepy;
	  maxi=ci+(vbe_height/4)*stepy;
	  break;
	}
      if(button&2)
	{
	  minr-=(x-vbe_width/4)*stepx*2;
	  maxr=minr+vbe_width*stepx*2;
	  mini-=(y-vbe_height/4)*stepy*2;
	  maxi=mini+vbe_height*stepy*2;
	  break;
	}
    }
}

void
main (int argc, char **argv)
{
  int x,mode=0x101;
  static double secs, c[4];
  clock_t vor, nach;
  clock_t zeit;
  char option,*colormap=NULL,s[32];

  c[0]=minr; c[1]=maxr; c[3]=-mini; c[2]=-maxi;
  while((option=getopt(argc,argv,"vi:c:m:"))!=EOF)
    switch(option)
      {
      case 'v':
	verbose=TRUE;
	printf("Boundary Tracing Demo V0.02, (C) 1997 Michael R. Ganss\n");
	break;
      case 'i':
	anz= atoi(optarg);
	break;
      case 'c':
	colormap=strdup(optarg);
	break;
      case 'm':
	x=atoi(optarg);
	switch(x)
	  {
	  case 1:
	    mode=0x101; /* 640x480x256 */
	    break;
	  case 2:
	    mode=0x103; /* 800x600x256 */
	    break;
	  case 3:
	    mode=0x105; /* 1024x768x256 */
	    break;
	  default:
	    break;
	  }
	break;
      case '?':
      default:
	if(!verbose)
	  fprintf(stderr,"Boundary Tracing Demo V0.02, (C) 1997 Michael R. Ganss\n");
	usage(argv);
      }
  argc-=optind;
  argv+=optind;
  x=0;
  while(argc&&(x<4))
    {
      c[x++]= atof(*argv);
      argc--;
      argv++;
    }
  minr=c[0]; maxr=c[1]; mini=-c[3]; maxi=-c[2];

  init_graphics(mode);
  init_mouse();
  /*show_cursor();*/
  set_mouselimits(vbe_width-1,vbe_height-1);
  
  if(colormap!=NULL)
    LoadColorMap(colormap);

  maxx=vbe_width; maxy=vbe_height;
  
  screen=(int **)malloc(maxy*sizeof(int*));
  for(y=0; y<maxy; y++)
    screen[y]=(int *)malloc(maxx*sizeof(int));
  
  while(1)
    {
      clearscreen(0);

      vor=clock();
      
      stepx = (maxr - minr) / maxx;
      stepy = (maxi - mini) / maxy;
      equal = (stepx < stepy ? stepx : stepy);

      for(y=0; y<maxy; y++)
	memset(screen[y],-1,maxx*sizeof(int));

      miny=0;
      if(mini<0&&maxi>0)
	{
	  midy=(int)(-mini/stepy+0.5);
	  maxy=midy;
	  for (y=0, basy=mini; y<maxy; y++, basy+=stepy)
	    for(x=0, basx=minr; x<maxx; x++, basx+=stepx)
	      if(screen[y][x]<0)
		trace(x,y,basx,basy);
	  
	  miny=midy;
	  maxy=vbe_height;
	  for (; y<maxy; y++, basy+=stepy)
	    for(x=0, basx=minr; x<maxx; x++, basx+=stepx)
	      if(screen[y][x]<0)
		trace(x,y,basx,basy);
	}
      else
	for (y=0, basy=mini; y<maxy; y++, basy+=stepy)
	  for(x=0, basx=minr; x<maxx; x++, basx+=stepx)
	    if(screen[y][x]<0)
	      trace(x,y,basx,basy);
      
      nach=clock();
	
      zeit = nach - vor;
      secs = fmod ((double) zeit / CLK_TCK, 60.0);
      if(verbose)
	{
	  sprintf(s,"%.2g secs", secs);
	  puttext(2,1,s,15);
	}
	  
      Zoom();
    }
}
