/* 
 * soi.c --
 *
 *	Simultaneous Orbit Iteration Image Generation Method. Computes
 *      rectangular regions by tracking the orbits of only a few key points.
 *
 * Copyright (c) 1994-1997 Michael R. Ganss. All Rights Reserved.
 *
 * This file is distributed under the same conditions as
 * AlmondBread. For further information see
 * <URL:http://www.cs.tu-berlin.de/~rms/AlmondBread>. 
 *
 */
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#undef WORDS_BIGENDIAN
#include "ieee754.h"
#include "vbe.h"
#include "default.h"

#define TRUE 1
#define FALSE 0
#define EVERY 15
#define BASIN_COLOR 0
#define INTEL

static double stepx, stepy;
static double minr=-2.5, maxr=1.5, mini=-1.5, maxi=1.5;
static double tolerance=0.1, twidth;
static float equal;
static int anz=10000;
static char basin = FALSE;
static int maxx, maxy;
static int verbose=FALSE;
extern int optind,opterr;
extern char *optarg;

static unsigned char adjust[256]={0,1,2,2,3,3,3,3,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8};
				  
int iteration(register double cr, register double ci,
	      register double re, register double im,
	      int start)
{
  register int iter,offset=0,k,n;
  register double ren,imn,sre,sim;
#ifdef INTEL
  register float mag;
  register unsigned long bail=0x41800000,magi; /* bail=16.0 */
#if 0 /* see below for explanation */
  register unsigned long eq=*(unsigned long *)&equal;
#endif
#else
  register double mag;
#endif
  union ieee754_double d;
  int exponent;
  
  if(basin)
    {
      sre=re; sim=im;
      ren=re*re;
      imn=im*im;
      if(start!=0)
	{
	  offset=anz-start+7;
	  iter=offset>>3;
	  offset&=7;
	  offset=(8-offset);
	}
      else
	iter=anz>>3;
      k=n=8;
      
      do
	{
	  im=im*re;
	  re=ren-imn;
	  im+=im;
	  re+=cr;
	  im+=ci;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;
	  
#if 0 /* this doesn't work properly, because integer compares are only
	 32 bits */
	  mag=fabs(sre-re);
	  magi=*(unsigned long *)&mag;
	  if(magi<eq)
	    {
	      mag=fabs(sim-im);
	      magi=*(unsigned long *)&mag;
	      if(magi<eq)
		return BASIN_COLOR;
	    }
#else
	  if(fabs(sre-re)<equal&&fabs(sim-im)<equal)
	    return BASIN_COLOR;
#endif

	  k-=8;
	  if(k<=0)
	    {
	      n<<=1;
	      sre=re;
	      sim=im;
	      k=n;
	    }

	  imn=im*im;
	  ren=re*re;
	  mag=ren+imn;
#ifdef INTEL
	  magi=*(unsigned long *)&mag;
#endif
	}
#ifdef INTEL
      while (magi<bail&&--iter!=0);
#else
      while (mag<16.0&&--iter!=0);
#endif
    }
  else
    {
      ren=re*re;
      imn=im*im;
      if(start!=0)
	{
	  offset=anz-start+7;
	  iter=offset>>3;
	  offset&=7;
	  offset=(8-offset);
	}
      else
	iter=anz>>3;

      do
	{      
	  im=im*re;
	  re=ren-imn;
	  im+=im;
	  re+=cr;
	  im+=ci;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*re;
	  ren=re+im;
	  re=re-im;
	  imn+=imn;
	  re=ren*re;
	  im=imn+ci;
	  re+=cr;

	  imn=im*im;
	  ren=re*re;
	  mag=ren+imn;
#ifdef INTEL
	  magi=*(unsigned long *)&mag;
#endif
	}
#ifdef INTEL
      while (magi<bail&&--iter!=0);
#else
      while (mag<16.0&&--iter!=0);
#endif
    }
  
  if(iter==0)
    {
      basin=TRUE;
      return BASIN_COLOR;
    }
  else
    {
      basin=FALSE;
#ifdef INTEL
      d.d=ren+imn;
#else
      d.d=mag;
#endif
      exponent=(d.ieee.exponent-_IEEE754_DOUBLE_BIAS)>>3;

      return (anz+offset-(((iter-1)<<3)+adjust[exponent]));
    }
}

static inline void putbox(int x1, int y1, int x2, int y2, int color)
{
  for(; y1<=y2; y1++)
    puthline(x1,y1,x2,color);
}

/* maximum side length beyond which we start regular scanning instead of
   subdividing */
#define SCAN 16

/* pixel interleave used in scanning */
#define INTERLEAVE 4

/* compute the value of the interpolation polynomial at (x,y) */
#define GET_REAL(x,y) \
interpolate(cim1,midi,cim2,\
	    interpolate(cre1,midr,cre2,zre1,zre5,zre2,x),\
	    interpolate(cre1,midr,cre2,zre6,zre9,zre7,x),\
	    interpolate(cre1,midr,cre2,zre3,zre8,zre4,x),y)
#define GET_IMAG(x,y) \
interpolate(cre1,midr,cre2,\
	    interpolate(cim1,midi,cim2,zim1,zim6,zim3,y),\
	    interpolate(cim1,midi,cim2,zim5,zim9,zim8,y),\
	    interpolate(cim1,midi,cim2,zim2,zim7,zim4,y),x)
     
/* compute the value of the interpolation polynomial at (x,y)
   from saved values before interpolation failed to stay within tolerance */
#define GET_SAVED_REAL(x,y) \
interpolate(cim1,midi,cim2,\
	    interpolate(cre1,midr,cre2,sr1,sr5,sr2,x),\
	    interpolate(cre1,midr,cre2,sr6,sr9,sr7,x),\
	    interpolate(cre1,midr,cre2,sr3,sr8,sr4,x),y)
#define GET_SAVED_IMAG(x,y) \
interpolate(cre1,midr,cre2,\
	    interpolate(cim1,midi,cim2,si1,si6,si3,y),\
	    interpolate(cim1,midi,cim2,si5,si9,si8,y),\
	    interpolate(cim1,midi,cim2,si2,si7,si4,y),x)
     
/* compute the value of the interpolation polynomial at (x,y)
   during scanning. Here, key values do not change, so we can precompute
   coefficients in one direction and simply evaluate the polynomial
   during scanning. */
#define GET_SCAN_REAL(x,y) \
interpolate(cim1,midi,cim2,\
	    EVALUATE(cre1,midr,br10,br11,br12,x),\
	    EVALUATE(cre1,midr,br20,br21,br22,x),\
	    EVALUATE(cre1,midr,br30,br31,br32,x),y)
#define GET_SCAN_IMAG(x,y) \
interpolate(cre1,midr,cre2,\
	    EVALUATE(cim1,midi,bi10,bi11,bi12,y),\
	    EVALUATE(cim1,midi,bi20,bi21,bi22,y),\
	    EVALUATE(cim1,midi,bi30,bi31,bi32,y),x)

/* compute coefficients of Newton polynomial (b0,..,b2) from
   (x0,w0),..,(x2,w2). */
#define INTERPOLATE(x0,x1,x2,w0,w1,w2,b0,b1,b2) \
b0=w0;\
b1=(w1-w0)/(long double)(x1-x0);\
b2=((w2-w1)/(long double)(x2-x1)-b1)/(x2-x0)

/* evaluate Newton polynomial given by (x0,b0),(x1,b1) at x:=t */     
#define EVALUATE(x0,x1,b0,b1,b2,t) \
((b2*(t-x1)+b1)*(t-x0)+b0)
     
/* Newton Interpolation.
   It computes the value of the interpolation polynomial given by
   (x0,w0)..(x2,w2) at x:=t */
static inline double interpolate(double x0, double x1, double x2,
			  double w0, double w1, double w2,
			  double t)
{
  register double b0=w0,b1=w1,b2=w2,b;

  /*b0=(r0*b1-r1*b0)/(x1-x0);
  b1=(r1*b2-r2*b1)/(x2-x1);
  b0=(r0*b1-r2*b0)/(x2-x0);

  return (double)b0;*/
  b=(b1-b0)/(x1-x0);
  return (double)((((b2-b1)/(x2-x1)-b)/(x2-x0))*(t-x1)+b)*(t-x0)+b0;
  /*
  if(t<x1)
    return w0+((t-x0)/(long double)(x1-x0))*(w1-w0);
  else
    return w1+((t-x1)/(long double)(x2-x1))*(w2-w1);*/
}

/* SOICompute - Perform simultaneous orbit iteration for a given rectangle

   Input: cre1..cim2 : values defining the four corners of the rectangle
          x1..y2     : corresponding pixel values
	  zre1..zim9 : intermediate iterated values of the key points (key values)
	  
	  (cre1,cim1)               (cre2,cim1)
	  (zre1,zim1)  (zre5,zim5)  (zre2,zim2)
	       +------------+------------+
	       |            |            |
	       |            |            |
	  (zre6,zim6)  (zre9,zim9)  (zre7,zim7)
	       |            |            |
	       |            |            |
	       +------------+------------+
	  (zre3,zim3)  (zre8,zim8)  (zre4,zim4)
	  (cre1,cim2)               (cre2,cim2)

	  iter       : current number of iterations
	  */
   
static inline void rhombus(double cre1, double cre2, double cim1, double cim2,
			   int x1, int x2, int y1, int y2,
			   double zre1, double zim1,
			   double zre2, double zim2,
			   double zre3, double zim3,
			   double zre4, double zim4,
			   double zre5, double zim5,
			   double zre6, double zim6,
			   double zre7, double zim7,
			   double zre8, double zim8,
			   double zre9, double zim9,
			   int iter)
{
  /* center of rectangle */
  double midr=(cre1+cre2)/2,midi=(cim1+cim2)/2;
  /* used in scanning */
  int x,y,z,color,savecolor,savex,helpcolor;
  double re,im,restep,imstep,interstep,helpre;
  double zre,zim;
  /* interpolation coefficients */
  long double br10,br11,br12,br20,br21,br22,br30,br31,br32;
  long double bi10,bi11,bi12,bi20,bi21,bi22,bi30,bi31,bi32;
  /* ratio of interpolated test point to iterated one */
  double l1,l2;
  /* squares of key values */
  double rq1,iq1;
  double rq2,iq2;
  double rq3,iq3;
  double rq4,iq4;
  double rq5,iq5;
  double rq6,iq6;
  double rq7,iq7;
  double rq8,iq8;
  double rq9,iq9;
  /* saved values of key values */
  double sr1,si1,sr2,si2,sr3,si3,sr4,si4;
  double sr5,si5,sr6,si6,sr7,si7,sr8,si8,sr9,si9;
  /* key values for subsequent rectangles */
  double re10,re11,re12,re13,re14,re15,re16,re17,re18,re19,re20,re21;
  double im10,im11,im12,im13,im14,im15,im16,im17,im18,im19,im20,im21;
  double re91,re92,re93,re94,im91,im92,im93,im94;
  /* test points */
  double cr1,cr2;
  double ci1,ci2;
  double tzr1,tzi1,tzr2,tzi2,tzr3,tzi3,tzr4,tzi4;
  double trq1,tiq1,trq2,tiq2,trq3,tiq3,trq4,tiq4;
  /* number of iterations before SOI iteration cycle */
  int before;

  if(iter>anz)
    {
      putbox(x1,y1,x2,y2,0);
      return;
    }
  
  if(y2-y1<=SCAN)
    {
      /* finish up the image by scanning the rectangle */
    scan:
      INTERPOLATE(cre1,midr,cre2,zre1,zre5,zre2,br10,br11,br12);
      INTERPOLATE(cre1,midr,cre2,zre6,zre9,zre7,br20,br21,br22);
      INTERPOLATE(cre1,midr,cre2,zre3,zre8,zre4,br30,br31,br32);
      
      INTERPOLATE(cim1,midi,cim2,zim1,zim6,zim3,bi10,bi11,bi12);
      INTERPOLATE(cim1,midi,cim2,zim5,zim9,zim8,bi20,bi21,bi22);
      INTERPOLATE(cim1,midi,cim2,zim2,zim7,zim4,bi30,bi31,bi32);
      
      restep=(cre2-cre1)/(x2-x1);
      imstep=(cim2-cim1)/(y2-y1);
      interstep=INTERLEAVE*restep;
      
      for(y=y1, im=cim1; y<y2; y++, im+=imstep)
	{
	  zre=GET_SCAN_REAL(cre1,im);
	  zim=GET_SCAN_IMAG(cre1,im);
	  savecolor=iteration(cre1,im,zre,zim,iter);
	  savex=x1;
	  for(x=x1+INTERLEAVE, re=cre1+interstep; x<x2;
	      x+=INTERLEAVE, re+=interstep)
	    {
	      zre=GET_SCAN_REAL(re,im);
	      zim=GET_SCAN_IMAG(re,im);
	      
	      if((color=iteration(re,im,zre,zim,iter))==savecolor)
		continue;
	      
	      for (z=x-1, helpre=re-restep; z>x-INTERLEAVE; z--,helpre-=restep)
		{
		  zre=GET_SCAN_REAL(helpre,im);
		  zim=GET_SCAN_IMAG(helpre,im);
		  if((helpcolor=iteration(helpre,im,zre,zim,iter))==savecolor)
		    break;
		  
		  putpixel(z,y,helpcolor&255);
		}
	      
	      if(savex<z)
		puthline(savex, y, z, savecolor&255);
	      else
		putpixel(savex, y, savecolor&255);
	      
	      savex = x;
	      savecolor = color;
	    }
	  
	  for (z=x2-1, helpre=cre2-restep; z>savex; z--,helpre-=restep)
	    {
	      zre=GET_SCAN_REAL(helpre,im);
	      zim=GET_SCAN_IMAG(helpre,im);
	      if((helpcolor=iteration(helpre,im,zre,zim,iter))==savecolor)
		break;
	      
	      putpixel(z,y,helpcolor&255);
	    }
	  
	  if(savex<z)
	    puthline(savex, y, z, savecolor&255);
	  else
	    putpixel(savex, y, savecolor&255);
	}
      return;
    }

  rq1=zre1*zre1; iq1=zim1*zim1;
  rq2=zre2*zre2; iq2=zim2*zim2;
  rq3=zre3*zre3; iq3=zim3*zim3;
  rq4=zre4*zre4; iq4=zim4*zim4;
  rq5=zre5*zre5; iq5=zim5*zim5;
  rq6=zre6*zre6; iq6=zim6*zim6;
  rq7=zre7*zre7; iq7=zim7*zim7;
  rq8=zre8*zre8; iq8=zim8*zim8;
  rq9=zre9*zre9; iq9=zim9*zim9;

  cr1=0.75*cre1+0.25*cre2; cr2=0.25*cre1+0.75*cre2;
  ci1=0.75*cim1+0.25*cim2; ci2=0.25*cim1+0.75*cim2;

  tzr1=GET_REAL(cr1,ci1);
  tzi1=GET_IMAG(cr1,ci1);

  tzr2=GET_REAL(cr2,ci1);
  tzi2=GET_IMAG(cr2,ci1);

  tzr3=GET_REAL(cr1,ci2);
  tzi3=GET_IMAG(cr1,ci2);

  tzr4=GET_REAL(cr2,ci2);
  tzi4=GET_IMAG(cr2,ci2);

  trq1=tzr1*tzr1;
  tiq1=tzi1*tzi1;

  trq2=tzr2*tzr2;
  tiq2=tzi2*tzi2;

  trq3=tzr3*tzr3;
  tiq3=tzi3*tzi3;

  trq4=tzr4*tzr4;
  tiq4=tzi4*tzi4;
  
  before=iter;
  
  do
    {
      sr1=zre1; si1=zim1;
      sr2=zre2; si2=zim2;
      sr3=zre3; si3=zim3;
      sr4=zre4; si4=zim4;
      sr5=zre5; si5=zim5;
      sr6=zre6; si6=zim6;
      sr7=zre7; si7=zim7;
      sr8=zre8; si8=zim8;
      sr9=zre9; si9=zim9;

      /* iterate key values */
      zim1=(zim1+zim1)*zre1+cim1;
      zre1=rq1-iq1+cre1;
      rq1=zre1*zre1;
      iq1=zim1*zim1;
      
      zim2=(zim2+zim2)*zre2+cim1;
      zre2=rq2-iq2+cre2;
      rq2=zre2*zre2;
      iq2=zim2*zim2;
      
      zim3=(zim3+zim3)*zre3+cim2;
      zre3=rq3-iq3+cre1;
      rq3=zre3*zre3;
      iq3=zim3*zim3;
      
      zim4=(zim4+zim4)*zre4+cim2;
      zre4=rq4-iq4+cre2;
      rq4=zre4*zre4;
      iq4=zim4*zim4;

      zim5=(zim5+zim5)*zre5+cim1;
      zre5=rq5-iq5+midr;
      rq5=zre5*zre5;
      iq5=zim5*zim5;
      
      zim6=(zim6+zim6)*zre6+midi;
      zre6=rq6-iq6+cre1;
      rq6=zre6*zre6;
      iq6=zim6*zim6;
      
      zim7=(zim7+zim7)*zre7+midi;
      zre7=rq7-iq7+cre2;
      rq7=zre7*zre7;
      iq7=zim7*zim7;
      
      zim8=(zim8+zim8)*zre8+cim2;
      zre8=rq8-iq8+midr;
      rq8=zre8*zre8;
      iq8=zim8*zim8;
      
      zim9=(zim9+zim9)*zre9+midi;
      zre9=rq9-iq9+midr;
      rq9=zre9*zre9;
      iq9=zim9*zim9;

      /* iterate test point */
      tzi1=(tzi1+tzi1)*tzr1+ci1;
      tzr1=trq1-tiq1+cr1;
      trq1=tzr1*tzr1;
      tiq1=tzi1*tzi1;

      tzi2=(tzi2+tzi2)*tzr2+ci1;
      tzr2=trq2-tiq2+cr2;
      trq2=tzr2*tzr2;
      tiq2=tzi2*tzi2;

      tzi3=(tzi3+tzi3)*tzr3+ci2;
      tzr3=trq3-tiq3+cr1;
      trq3=tzr3*tzr3;
      tiq3=tzi3*tzi3;

      tzi4=(tzi4+tzi4)*tzr4+ci2;
      tzr4=trq4-tiq4+cr2;
      trq4=tzr4*tzr4;
      tiq4=tzi4*tzi4;

      iter++;

      /* if one of the iterated values bails out, subdivide */
      if((rq1+iq1)>16.0||
	 (rq2+iq2)>16.0||
	 (rq3+iq3)>16.0||
	 (rq4+iq4)>16.0||
	 (rq5+iq5)>16.0||
	 (rq6+iq6)>16.0||
	 (rq7+iq7)>16.0||
	 (rq8+iq8)>16.0||
	 (rq9+iq9)>16.0||
	 (trq1+tiq1)>16.0||
	 (trq2+tiq2)>16.0||
	 (trq3+tiq3)>16.0||
	 (trq4+tiq4)>16.0)
	break;

      /* if maximum number of iterations is reached, the whole rectangle
	 can be assumed part of M. This is of course best case behavior
	 of SOI, we seldomly get there */
      if(iter>anz)
	{
	  putbox(x1,y1,x2,y2,0);
	  return;
	}

      /* now for all test points, check whether they exceed the
	 allowed tolerance. if so, subdivide */
      l1=GET_REAL(cr1,ci1);
      l1=(tzr1==0.0)?
	(l1==0.0)?1.0:1000.0:
	l1/tzr1;
      if(fabs(1.0-l1)>twidth)
	break;
      
      l2=GET_IMAG(cr1,ci1);
      l2=(tzi1==0.0)?
	(l2==0.0)?1.0:1000.0:
	l2/tzi1;
      if(fabs(1.0-l2)>twidth)
	break;
      
      l1=GET_REAL(cr2,ci1);
      l1=(tzr2==0.0)?
	(l1==0.0)?1.0:1000.0:
	l1/tzr2;
      if(fabs(1.0-l1)>twidth)
	break;

      l2=GET_IMAG(cr2,ci1);
      l2=(tzi2==0.0)?
	(l2==0.0)?1.0:1000.0:
	l2/tzi2;
      if(fabs(1.0-l2)>twidth)
	break;

      l1=GET_REAL(cr1,ci2);
      l1=(tzr3==0.0)?
	(l1==0.0)?1.0:1000.0:
	l1/tzr3;
      if(fabs(1.0-l1)>twidth)
	break;
      
      l2=GET_IMAG(cr1,ci2);
      l2=(tzi3==0.0)?
	(l2==0.0)?1.0:1000.0:
	l2/tzi3;
      if(fabs(1.0-l2)>twidth)
	break;
      
      l1=GET_REAL(cr2,ci2);
      l1=(tzr4==0.0)?
	(l1==0.0)?1.0:1000.0:
	l1/tzr4;
      if(fabs(1.0-l1)>twidth)
	break;

      l2=GET_IMAG(cr2,ci2);
      l2=(tzi4==0.0)?
	(l2==0.0)?1.0:1000.0:
	l2/tzi4;
      if(fabs(1.0-l2)>twidth)
	break;
    }
  while(1);

  iter--;

  /* this is a little heuristic I tried to improve performance. */
  if(iter-before<10)
    {
      zre1=sr1; zim1=si1;
      zre2=sr2; zim2=si2;
      zre3=sr3; zim3=si3;
      zre4=sr4; zim4=si4;
      zre5=sr5; zim5=si5;
      zre6=sr6; zim6=si6;
      zre7=sr7; zim7=si7;
      zre8=sr8; zim8=si8;
      zre9=sr9; zim9=si9;
      goto scan;
    }

  /* compute key values for subsequent rectangles */
  
  re10=interpolate(cre1,midr,cre2,sr1,sr5,sr2,cr1);
  im10=interpolate(cre1,midr,cre2,si1,si5,si2,cr1);

  re11=interpolate(cre1,midr,cre2,sr1,sr5,sr2,cr2);
  im11=interpolate(cre1,midr,cre2,si1,si5,si2,cr2);

  re20=interpolate(cre1,midr,cre2,sr3,sr8,sr4,cr1);
  im20=interpolate(cre1,midr,cre2,si3,si8,si4,cr1);

  re21=interpolate(cre1,midr,cre2,sr3,sr8,sr4,cr2);
  im21=interpolate(cre1,midr,cre2,si3,si8,si4,cr2);

  re15=interpolate(cre1,midr,cre2,sr6,sr9,sr7,cr1);
  im15=interpolate(cre1,midr,cre2,si6,si9,si7,cr1);

  re16=interpolate(cre1,midr,cre2,sr6,sr9,sr7,cr2);
  im16=interpolate(cre1,midr,cre2,si6,si9,si7,cr2);
  
  re12=interpolate(cim1,midi,cim2,sr1,sr6,sr3,ci1);
  im12=interpolate(cim1,midi,cim2,si1,si6,si3,ci1);

  re14=interpolate(cim1,midi,cim2,sr2,sr7,sr4,ci1);
  im14=interpolate(cim1,midi,cim2,si2,si7,si4,ci1);

  re17=interpolate(cim1,midi,cim2,sr1,sr6,sr3,ci2);
  im17=interpolate(cim1,midi,cim2,si1,si6,si3,ci2);

  re19=interpolate(cim1,midi,cim2,sr2,sr7,sr4,ci2);
  im19=interpolate(cim1,midi,cim2,si2,si7,si4,ci2);

  re13=interpolate(cim1,midi,cim2,sr5,sr9,sr8,ci1);
  im13=interpolate(cim1,midi,cim2,si5,si9,si8,ci1);

  re18=interpolate(cim1,midi,cim2,sr5,sr9,sr8,ci2);
  im18=interpolate(cim1,midi,cim2,si5,si9,si8,ci2);

  re91=GET_SAVED_REAL(cr1,ci1);
  re92=GET_SAVED_REAL(cr2,ci1);
  re93=GET_SAVED_REAL(cr1,ci2);
  re94=GET_SAVED_REAL(cr2,ci2);

  im91=GET_SAVED_IMAG(cr1,ci1);
  im92=GET_SAVED_IMAG(cr2,ci1);
  im93=GET_SAVED_IMAG(cr1,ci2);
  im94=GET_SAVED_IMAG(cr2,ci2);
  
  rhombus(cre1,midr,cim1,midi,
	  x1,(x1+x2)>>1,y1,(y1+y2)>>1,
	  sr1,si1,
	  sr5,si5,
	  sr6,si6,
	  sr9,si9,
	  re10,im10,
	  re12,im12,
	  re13,im13,
	  re15,im15,
	  re91,im91,
	  iter);
  rhombus(midr,cre2,cim1,midi,
	  (x1+x2)>>1,x2,y1,(y1+y2)>>1,
	  sr5,si5,
	  sr2,si2,
	  sr9,si9,
	  sr7,si7,
	  re11,im11,
	  re13,im13,
	  re14,im14,
	  re16,im16,
	  re92,im92,
	  iter);
  rhombus(cre1,midr,midi,cim2,
	  x1,(x1+x2)>>1,(y1+y2)>>1,y2,
	  sr6,si6,
	  sr9,si9,
	  sr3,si3,
	  sr8,si8,
	  re15,im15,
	  re17,im17,
	  re18,im18,
	  re20,im20,
	  re93,im93,
	  iter);
  rhombus(midr,cre2,midi,cim2,
	  (x1+x2)>>1,x2,(y1+y2)>>1,y2,
	  sr9,si9,
	  sr7,si7,
	  sr8,si8,
	  sr4,si4,
	  re16,im16,
	  re18,im18,
	  re19,im19,
	  re21,im21,
	  re94,im94,
	  iter);
}

void LoadColorMap(char *filename)
{
  FILE *f;
  int n=256,index,r,g,b;
  char c;

  if((f=fopen(filename,"r"))==NULL)
    {
      perror(filename);
      close_graphics();
      exit(1);
    }
  for(index=0; index<n; index++)
    {
      if(fscanf(f,"%d",&r)==EOF)
	{
	  if(ferror(f))
	    perror(filename);
	  else fprintf(stderr,"%s: Premature EOF.\n",filename);
	  close_graphics();
	  exit(1);
	}
      if(fscanf(f,"%d",&g)==EOF)
	{
	  if(ferror(f))
	    perror(filename);
	  else fprintf(stderr,"%s: Premature EOF.\n",filename);
	  close_graphics();
	  exit(1);
	}
      if(fscanf(f,"%d",&b)==EOF)
	{
	  if(ferror(f))
	    perror(filename);
	  else fprintf(stderr,"%s: Premature EOF.\n",filename);
	  close_graphics();
	  exit(1);
	}
      while(((c=fgetc(f))!='\n')&&(c!=EOF));
      set_palette(index,(int)((double)r/4+0.5),
		  (int)((double)g/4+0.5),(int)((double)b/4+0.5));
    }
  fclose(f);
}

void usage(char **argv)
{
  int i,n;
  
  fprintf(stderr,"Usage: %s %n[-v]\n",argv[0],&n);
  for(i=0; i<n; i++)
    fprintf(stderr," ");
  fprintf(stderr,"[-i max # of iterations]\n");
  for(i=0; i<n; i++)
    fprintf(stderr," ");
  fprintf(stderr,"[-c colormap]\n");
  for(i=0; i<n; i++)
    fprintf(stderr," ");
  fprintf(stderr,"[-m graphics mode 1|2|3]\n");
  for(i=0; i<n; i++)
    fprintf(stderr," ");
  fprintf(stderr,"[-t tolerance]\n");
  for(i=0; i<n; i++)
    fprintf(stderr," ");
  fprintf(stderr,"-- [minr] [maxr] [mini] [maxi]\n");
  
  exit(1);
}

void Zoom(void)
{
  int x,y,button;
  double cr,ci;

  vbe_cursor=0;
  
  while(1)
    {
      mouse_position(&x,&y,&button);
      display_cursor(x,y);
      if(button&1)
	{
	  cr=minr+x*stepx;
	  ci=mini+y*stepy;
	  minr=cr-(vbe_width/4)*stepx;
	  maxr=cr+(vbe_width/4)*stepx;
	  mini=ci-(vbe_height/4)*stepy;
	  maxi=ci+(vbe_height/4)*stepy;
	  break;
	}
      if(button&2)
	{
	  minr-=(x-vbe_width/4)*stepx*2;
	  maxr=minr+vbe_width*stepx*2;
	  mini-=(y-vbe_height/4)*stepy*2;
	  maxi=mini+vbe_height*stepy*2;
	  break;
	}
    }
}

void
main (int argc, char **argv)
{
  int x,mode=0x101;
  static double secs;
  static double c[4]={-0.1049775673947832,-0.1049775673813040,
		      0.9270413637758836,0.9270413637859878,};
  clock_t vor, nach;
  clock_t zeit;
  char option,*colormap=NULL,s[32];

  while((option=getopt(argc,argv,"vi:c:m:t:"))!=EOF)
    switch(option)
      {
      case 'v':
	verbose=TRUE;
	printf("SOI Demo V0.02, (C) 1997 Michael R. Ganss\n");
	break;
      case 'i':
	anz= atoi(optarg);
	break;
      case 'c':
	colormap=strdup(optarg);
	break;
      case 'm':
	x=atoi(optarg);
	switch(x)
	  {
	  case 1:
	    mode=0x101; /* 640x480x256 */
	    break;
	  case 2:
	    mode=0x103; /* 800x600x256 */
	    break;
	  case 3:
	    mode=0x105; /* 1024x768x256 */
	    break;
	  default:
	    break;
	  }
	break;
      case 't':
	tolerance=atof(optarg);
	break;
      case '?':
      default:
	if(!verbose)
	  fprintf(stderr,"SOI Demo V0.02, (C) 1997 Michael R. Ganss\n");
	usage(argv);
      }
  argc-=optind;
  argv+=optind;
  x=0;
  while(argc&&(x<4))
    {
      c[x++]= atof(*argv);
      argc--;
      argv++;
    }
  minr=c[0]; maxr=c[1]; mini=-c[3]; maxi=-c[2];

  init_graphics(mode);
  init_mouse();
  /*show_cursor();*/
  set_mouselimits(vbe_width-1,vbe_height-1);
  
  if(colormap!=NULL)
    LoadColorMap(colormap);

  maxx=vbe_width; maxy=vbe_height;
  twidth=tolerance/vbe_width;
  
  while(1)
    {
      clearscreen(0);

      vor=clock();
      
      stepx = (maxr - minr) / maxx;
      stepy = (maxi - mini) / maxy;
      equal = (stepx < stepy ? stepx : stepy);

      rhombus(minr,maxr,mini,maxi,
	      0,maxx,0,maxy,
	      minr,mini,
	      maxr,mini,
	      minr,maxi,
	      maxr,maxi,
	      (maxr+minr)/2,mini,
	      minr,(mini+maxi)/2,
	      maxr,(mini+maxi)/2,
	      (maxr+minr)/2,maxi,
	      (minr+maxr)/2,(mini+maxi)/2,
	      1);
      
      nach=clock();
	
      zeit = nach - vor;
      secs = fmod ((double) zeit / CLK_TCK, 60.0);
      if(verbose)
	{
	  sprintf(s,"%.2g secs", secs);
	  puttext(2,1,s,15);
	}
	  
      Zoom();
    }
}
