/* 
 * vbe.h --
 *
 * VESA BE 2.0 graphics routines.
 *
 * Adapted from vbetest.c, which is
 * Copyright (c) Charles W Sandmann (sandmann@clio.rice.edu).
 * All other code is
 * Copyright (c) 1997 Michael R. Ganss. All Rights Reserved.
 *
 * This file is distributed under the same conditions as
 * AlmondBread. For further information see
 * <URL:http://www.cs.tu-berlin.de/~rms/AlmondBread>. 
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pc.h>
#include <dpmi.h>
#include <go32.h>
#include <dos.h>

#include <sys/nearptr.h>
#include <crt0.h>
int _crt0_startup_flags = _CRT0_FLAG_NEARPTR | _CRT0_FLAG_NONMOVE_SBRK;

#define BYTE unsigned char

typedef struct {
  unsigned ModeAttributes;
  unsigned granularity,startseg;
  __dpmi_raddr func;
  short bscanline;
  short XResolution;
  short YResolution;
  short charpixels;
  unsigned bogus1,bogus2,bogus3,bogus4;
  unsigned PhysBasePtr;
  char bogus[228];
} ModeInfoBlock;

unsigned int vbe_addr;
int vbe_width,vbe_height,vbe_linear=0,vbe_bank=0;
int *vbe_offset;
BYTE *vbe_video;

ModeInfoBlock *get_mode_info(int mode)
{
  static ModeInfoBlock info;
  __dpmi_regs r;

  /* Use the transfer buffer to store the results of VBE call */
  r.x.ax = 0x4F01;
  r.x.cx = mode;
  r.x.es = __tb / 16;
  r.x.di = 0;
  __dpmi_int(0x10, &r);
  if(r.h.ah) return 0;
  dosmemget(__tb, sizeof(ModeInfoBlock), &info);
  return &info;
}

void set_palette(int entry,int r,int g,int b)
{
  outportb(0x3c8,(BYTE)entry);
  outportb(0x3c9,(BYTE)r);
  outportb(0x3c9,(BYTE)g);
  outportb(0x3c9,(BYTE)b);
}

void init_graphics(int mode)
{
  __dpmi_meminfo info;
  __dpmi_regs reg;
  ModeInfoBlock *mb;
  int y;

  mb=get_mode_info(mode);

  if (!mb) {
    printf("Get VESA mode info failed.\n");
    exit(1);
  }
  if ((mb->ModeAttributes & 0x80)) {
    vbe_linear=1;
    printf("Using linear frame buffer.\n");
  }
  else
    printf("Using banked mode.\n");

  vbe_width=mb->XResolution;
  vbe_height=mb->YResolution;
  if(vbe_linear)
    {
      info.size=vbe_width*vbe_height;
      info.address=mb->PhysBasePtr;
    }
  else
    {
      info.size=65536;
      info.address=(unsigned int)0xA0000;
    }
  if(__dpmi_physical_address_mapping(&info) == -1) {
    printf("Physical mapping of address 0x%x failed!\n",mb->PhysBasePtr);
    exit(2);
  }
  vbe_addr=info.address;		/* Updated by above call */
  if(vbe_linear)
    printf("Vesa mode %x: %d x %d, linear frame: 0x%x\n", mode,
	   vbe_width,vbe_height,vbe_addr);
  else
    printf("Vesa mode %x: %d x %d, banked frame: 0x%x\n", mode,
	   vbe_width,vbe_height,vbe_addr);
  delay(1500);

  reg.x.ax=0x4f02;
  if(vbe_linear)
    reg.x.bx=mode|0x4000;		/* mode plus linear enable bit */
  else
    reg.x.bx=mode;
  __dpmi_int(0x10,&reg);		/* set the mode */
  if(reg.h.al != 0x4f || reg.h.ah) {
    printf("Mode set failed!\n");
    exit(3);
  }
  vbe_video = (char *)(vbe_addr + __djgpp_conventional_base);
  vbe_offset=(int *)malloc(vbe_height*sizeof(int));
  for(y=0; y<vbe_height; y++)
    vbe_offset[y]=y*vbe_width;
}

void close_graphics(void)
{
  __dpmi_regs reg;
  reg.x.ax = 0x0003;
  __dpmi_int(0x10,&reg);
}

static inline void setbank(int bank)
{
  __dpmi_regs reg;
  
  reg.x.ax=0x4f05;
  reg.x.bx=0;
  reg.x.dx=bank;
  __dpmi_int(0x10,&reg);
  vbe_bank=bank;
}

static inline void putpixel(int x, int y, int c)
{
  int pixel,bank;

  if(vbe_linear)
    vbe_video[vbe_offset[y]+x]=(BYTE)c;
  else
    {
      pixel=vbe_offset[y]+x;
      bank=pixel>>16;
      if(bank!=vbe_bank)
	setbank(bank);
      vbe_video[pixel&65535]=(BYTE)c;
    }
}

static inline void putpixel_linear(int x, int y, int c)
{
  vbe_video[vbe_offset[y]+x]=(BYTE)c;
}

static inline void putpixel_banked(int x, int y, int c)
{
  int pixel,bank;

  pixel=vbe_offset[y]+x;
  bank=pixel>>16;
  if(bank!=vbe_bank)
    setbank(bank);
  vbe_video[pixel&65535]=(BYTE)c;
}

static inline void puthline(int x1, int y, int x2, int c)
{
  int bank1,bank2,pixel1,pixel2;
  
  if(vbe_linear)
    memset(&vbe_video[vbe_offset[y]+x1],c,x2-x1+1);
  else
    {
      pixel1=vbe_offset[y]+x1;
      pixel2=vbe_offset[y]+x2;
      bank1=pixel1>>16;
      bank2=pixel2>>16;
      if(bank1==bank2)
	{
	  if(bank1!=vbe_bank)
	    setbank(bank1);
	  memset(&vbe_video[pixel1&65535],c,x2-x1+1);
	}
      else
	{
	  if(bank1!=vbe_bank)
	    setbank(bank1);
	  pixel1&=65535;
	  memset(&vbe_video[pixel1],c,65536-pixel1);
	  setbank(bank2);
	  memset(vbe_video,c,(pixel2&65535)+1);
	}
    }
}

static inline void puthline_linear(int x1, int y, int x2, int c)
{
  memset(&vbe_video[vbe_offset[y]+x1],c,x2-x1+1);
}

static inline void puthline_banked(int x1, int y, int x2, int c)
{
  int bank1,bank2,pixel1,pixel2;
  
  pixel1=vbe_offset[y]+x1;
  pixel2=vbe_offset[y]+x2;
  bank1=pixel1>>16;
  bank2=pixel2>>16;
  if(bank1==bank2)
    {
      if(bank1!=vbe_bank)
	setbank(bank1);
      memset(&vbe_video[pixel1&65535],c,x2-x1+1);
    }
  else
    {
      if(bank1!=vbe_bank)
	setbank(bank1);
      pixel1&=65535;
      memset(&vbe_video[pixel1],c,65536-pixel1);
      setbank(bank2);
      memset(vbe_video,c,(pixel2&65535)+1);
    }
}
      
static inline void clearscreen(int c)
{
  int banks,i;
  
  if(vbe_linear)
    memset(vbe_video,(BYTE)c,vbe_width*vbe_height);
  else
    {
      banks=(vbe_width*vbe_height)>>16;
      for(i=0; i<banks; i++)
	{
	  setbank(i);
	  memset(vbe_video,(BYTE)c,65536);
	}
      setbank(banks);
      memset(vbe_video,(BYTE)c,(vbe_width*vbe_height)&65535);
    }
}

/* 12x16 pointer */
static BYTE pointer[] = {
  0,1,0,0,0,0,0,0,0,0,0,0,
  1,2,1,0,0,0,0,0,0,0,0,0,
  1,2,2,1,0,0,0,0,0,0,0,0,
  1,2,2,2,1,0,0,0,0,0,0,0,
  1,2,2,2,2,1,0,0,0,0,0,0,
  1,2,2,2,2,2,1,0,0,0,0,0,
  1,2,2,2,2,2,2,1,0,0,0,0,
  1,2,2,2,2,2,2,2,1,0,0,0,
  1,2,2,2,2,2,2,2,2,1,0,0,
  1,2,2,2,2,2,2,2,2,2,1,0,
  1,2,2,2,2,2,2,2,2,2,2,1,
  1,2,2,2,2,1,1,1,1,1,1,0,
  1,2,2,2,1,0,0,0,0,0,0,0,
  1,2,2,1,0,0,0,0,0,0,0,0,
  1,2,1,0,0,0,0,0,0,0,0,0,
  0,1,0,0,0,0,0,0,0,0,0,0,
};
static BYTE buffer[12*16];

static inline int getpixel(int x, int y)
{
  int pixel,bank;

  if(vbe_linear)
    return (int)vbe_video[vbe_offset[y]+x];
  else
    {
      pixel=vbe_offset[y]+x;
      bank=pixel>>16;
      if(bank!=vbe_bank)
	setbank(bank);
      return (int)vbe_video[pixel&65535];
    }
}

static int vbe_cursor=0,vbe_cx,vbe_cy;
static inline void display_cursor(int x, int y)
{
  int mx,my,bx,by,px,py,color;

  if(vbe_cursor&&x==vbe_cx&&y==vbe_cy)
    return;
  
  if(vbe_cursor)
    {
      mx=(vbe_cx+12>vbe_width?vbe_width:vbe_cx+12);
      my=(vbe_cy+16>vbe_height?vbe_height:vbe_cy+16);
      
      for(by=vbe_cy, py=0; by<my; py++, by++)
	for(bx=vbe_cx, px=0; bx<mx; px++, bx++)
	  putpixel(bx,by,buffer[py*12+px]);
    }
  
  mx=(x+12>vbe_width?vbe_width:x+12);
  my=(y+16>vbe_height?vbe_height:y+16);
  
  for(by=y, py=0; by<my; py++, by++)
    for(bx=x, px=0; bx<mx; px++, bx++)
      buffer[py*12+px]=(BYTE)getpixel(bx,by);

  for(by=y, py=0; by<my; py++, by++)
    for(bx=x, px=0; bx<mx; px++, bx++)
      if((color=pointer[py*12+px])!=0)
	putpixel(bx,by,color);

  vbe_cursor=1;
  vbe_cx=x;
  vbe_cy=y;
}
 
static inline void init_mouse(void)
{
  __dpmi_regs r;

  r.x.ax = 0;
  __dpmi_int(0x33, &r);
}

static inline void show_cursor(void)
{
  __dpmi_regs r;

  r.x.ax = 1;
  __dpmi_int(0x33, &r);
}

static inline void set_mouselimits(int x, int y)
{
  __dpmi_regs r;

  r.x.ax = 7;
  r.x.cx = 0;
  r.x.dx = x;
  __dpmi_int(0x33, &r);
  r.x.ax = 8;
  r.x.cx = 0;
  r.x.dx = y;
  __dpmi_int(0x33, &r);
}

static inline void mouse_position(int *x, int *y, int *button)
{
  __dpmi_regs r;

  r.x.ax = 3;
  __dpmi_int(0x33, &r);
  *x=r.x.cx;
  *y=r.x.dx;
  *button=r.x.bx;
}

static inline void get_mouselimits(int *x, int *y)
{
  __dpmi_regs r;

  r.x.ax = 0x26;
  __dpmi_int(0x33, &r);
  *x=r.x.cx;
  *y=r.x.dx;
}

static inline void disable_mouse(void)
{
  __dpmi_regs r;

  r.x.ax = 0x1f;
  __dpmi_int(0x33, &r);
}

static inline void enable_mouse(void)
{
  __dpmi_regs r;

  r.x.ax = 0x20;
  __dpmi_int(0x33, &r);
}
